;; Copyright (c) 1998 Franz Inc, Berkeley, CA  - All rights reserved.
;;
;; The software, data and information contained herein are proprietary
;; to, and comprise valuable trade secrets of, Franz, Inc.  They are
;; given in confidence by Franz, Inc. pursuant to a written license
;; agreement, and may be stored and used only in accordance with the terms
;; of such license.
;;
;; Restricted Rights Legend
;; ------------------------
;; Use, duplication, and disclosure of the software, data and information
;; contained herein by any agency, department or entity of the U.S.
;; Government are subject to restrictions of Restricted Rights for
;; Commercial Software developed at private expense as specified in
;; DOD FAR Supplement 52.227-7013 (c) (1) (ii), as applicable.
;;
;; $Id: checklinks.cl,v 1.1.2.13 1999/06/18 19:27:39 layer Exp $
;;
;; A program to check for broken links in HTML documents.  It recursively
;; descends into locally referenced file documents, too.  It keeps a table
;; of visited documents, so it won't get into a loop.

(in-package :user)

(eval-when (compile eval load) (require :aclwin))

#+ignore ;; causes the `check-links' call to be profiled.
(eval-when (compile) (pushnew :profile-check-links *features*))

(defvar *production-mode* t) ;; for debugging make this `nil'

(defvar *print-forms* nil)

(defvar *check-all-links* nil
  "When given the -a command line argument, verify even http: links.")

(defvar *usage*
    "~
Usage: checklinks [-D] [-d directory] file1 file2 ...
-a    connect to remote sites to check that links exists
-A alias1=value1,alias2=value2,...
      enter aliases for URLs (for Apache ScriptAliases)
-D    debug mode: print informative messages
      (all by itself, -d runs a primitive top-level)
-d directory
      check for orphaned files in directory, files that are not pointed to
      by any of the html files traversed in the given file<N> arguments
-p    debug mode: print the check-links form
-v    verbose output (print names of files as they are scanned)
-x file
      exclude file from scan
-X dir1,dir2,...
      directories to ignore
file1 file2 ...
      check the given filename arguments for bad links
")

(defun main ()
  (handler-case
      (system:with-command-line-arguments
	  ("aA:Dd:I:pvx:X:"
	   check-all aliases debug-mode directory image print-form verbose
	   ignore-files ignore-directories)
	  (files)
	(declare (ignore image))
	(when aliases
	  (when print-form (format t "~s~%" `(initialize-aliases ,aliases)))
	  (initialize-aliases aliases))
	(when debug-mode (setq *production-mode* nil))
	(when verbose (setq *print-forms* t))
	(when (null files)
	  (if* debug-mode
	     then (break "hack away, dude...")
	     else (format t *usage*)
		  (exit 0 :quiet t)))
	(let ((status 0))
	  (dolist (file files)
	    (let ((ignore-files 
		   (when ignore-files
		     (mapcar #'pathname
			     (aclwin:delimited-string-to-list
			      ignore-files
			      #\,))))
		  (ignore-directories
		   (when ignore-directories
		     (mapcar #'pathname
			     (aclwin:delimited-string-to-list
			      ignore-directories
			      #\,)))))
	      (when print-form
		(format t "~s~%"
			`(check-links ,file
				      :directory ,directory
				      :ignore-files ',ignore-files
				      :ignore-directories ',ignore-directories
				      :check-all ,check-all)))
	      (incf status
		    (check-links file
				 :directory directory
				 :ignore-files ignore-files
				 :ignore-directories ignore-directories
				 :check-all check-all))))
	  (exit status :quiet t)))
    (error (c)
      (if* *production-mode*
	 then (format t "An error occurred: ~a.~%" c)
	      (exit 1 :quiet t)
	 else (error c)))))

(defvar *image-file-types* '("gif" "jpg"))

(defvar *linked-files* nil
  "A list of files to visit, built up over the course of the recursion.")

(defvar *image-files*
    (make-hash-table :size 777 :test #+mswindows #'equalp #-mswindows #'equal)
  "A list of image files we've seen, built up over the course of the
recursion.")

(defvar *processed-files*
    ;; nil
    (make-hash-table :size 777 :test #+mswindows #'equalp #-mswindows #'equal)
  "A list of processed files, so we don't loop infinitely.")

(defvar *book-mark-table*
    (make-hash-table :size 101 :test #+mswindows #'equalp #-mswindows #'equal)
  "A list of per-file bookmarks to verify.")

(defvar *nonexistent-files* nil)
(defvar *root-dir* nil)

(defparameter *default-http-port* 80)

(defvar .printed-file-name.)

(declaim (optimize
	  ;; Some pretty intense string stuff going on here, so crank up
	  ;; the speed and down the safety.
	  (speed 3) (safety 0)))

#+profile-check-links
(eval-when (compile eval load) (require :prof))

(defun check-links (file
		    &key directory
			 (recurse t)
			 ignore-files
			 ignore-directories
			 ((:non-existent-files *nonexistent-files*) nil)
			 ((:check-all *check-all-links*) nil)
		    &aux (file (merge-pathnames file))
			 (*root-dir* (pathname (excl::path-namestring file)))
			 (status 0))
  "Check the HTML links in FILE.  Recursively descend into locally
referenced file documents, and check links there, too.  Keep a table 
of visited documents, so we won't get into a loop.  The :IGNORE argument
specifies bad links that we should ignore, and must be a list of filenames
without path elements.  The :DIRECTORY argument specifies a web root that
should additionally be checked: currently, files that are not pointed to by
any HTML document are highlighted."
  #+profile-check-links (prof:start-profiler)
  (setq *linked-files* nil)
  (clrhash *image-files*)
  (clrhash *book-mark-table*)
  (clrhash *processed-files*)
  ;; Pass 1: start with `file' checking links.
  (when directory
    (format t ";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; PASS 1~%"))
  (incf status (check-links-1 file
			      :recurse recurse :ignore-files ignore-files
			      :ignore-directories ignore-directories))

  (format t ";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; PASS 2~%")
  ;; process unprocessed bookmarks...
  (maphash
   #'(lambda (file bookmark-item-list)
       (when bookmark-item-list
	 (dolist (bookmark-item bookmark-item-list) 
	   (let ((bookmark (car bookmark-item))
		 (referencing-file (cdr bookmark-item)))
	     (when (null (find-bookmark-definition (file-contents file)
						   bookmark))
	       (note-bad-link
		file
		"~
Bookmark ~s does not have a definition in
   ~a
   yet it was referenced from file
   ~a.~%"
		bookmark
		(enough-namestring referencing-file *root-dir*)
		(enough-namestring file *root-dir*))
	       (incf status)
	       ;; so we don't see it again in pass 3:
	       (setf (gethash (namestring file) *book-mark-table*) nil))))))
   *book-mark-table*)

  (when (and recurse directory)
    ;; Pass 3: find files in `directory' we haven't process yet.  Warn the
    ;;         user about this, and process them for bad links.
    (format t ";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; PASS 3~%")
    (map-over-directory
     #'(lambda (p &aux (enough-p (enough-namestring p *root-dir*)))
	 (if* (and ignore-files
		   (member (file-namestring enough-p)
			   ignore-files
			   :test #'equalp :key #'namestring))
	    thenret
	  elseif (and ignore-directories
		      (member (directory-namestring enough-p)
			      ignore-directories
			      :test #'(lambda (a b) (match-regexp b a))
			      :key #'namestring))
	    thenret
	  elseif (member (pathname-type p) '("htm" "html") :test #'equalp)
	    then (when (not (visited-file-p p))
		   (format t "File ~a not pointed to by any file.~%" enough-p)
		   (incf status
			 (check-links-1
			  p
			  :recurse recurse
			  :ignore-files ignore-files
			  :ignore-directories ignore-directories)))
	  elseif (member (pathname-type p) *image-file-types*
			 :test #'equalp)
	    then (when (not (image-file-seen-p enough-p))
		   (format t "Image file ~a not pointed to by any file.~%"
			   enough-p))
	    else ;; ignore the file
		 nil))
     directory))
  (format t "Found ~d bad links.~%" status)
  #+profile-check-links
  (progn
    (prof:stop-profiler)
    (prof:show-call-graph)
    (prof:show-flat-profile))
  status)

(defparameter *link-regexp*
    (let ((spaces '(#\space #\tab #\newline #\return))
	  (string "\"\\([^\"]*\\)\""))
      (compile-regexp
       (concatenate 'simple-string
	 "\\("
;;;; html link:
	 "<a[" spaces "]+[^>]*href=" string "[^>]*>" "\\|"
;;;; frame:
	 "<frame[" spaces "]+[^>]*src=" string "[^>]*>" "\\|"
;;;; comment:
	 "<!--\\([^>]*\\)-->" "\\|"
;;;; an applet parameter:
	 "<param[" spaces "]+[^>]*name=\"url\"[^>]*value=" string "[^>]*>" "\\|"
;;;; image in a <body...>:
	 "<body[" spaces "]+[^>]*background=" string "[^>]*>" "\\|"
;;;; image link:
	 "<img[" spaces "]+[^>]*src=" string
	 "\\([^>]*[" spaces "]+[^>]*src=" string "\\|\\)" "[^>]*>"
	 "\\)"
	 ))))

(defparameter *http-ok-response*
    ;; 2xx is OK and 3xx is redirect
    (compile-regexp "HTTP/.\\.. [23].."))

(defun check-links-1 (current-file
		      &key (recurse t)
			   ignore-files
			   ignore-directories
		      &aux (status 0)
			   (.printed-file-name. nil))
  (flet
      ((process-match (positions html-buffer visible-file visible-position
		       include-map image-positions
		       &aux start end link)
	 (setq start (car positions))
	 (setq end (cdr positions))
	 (setq link (subseq html-buffer start end))
	
	 (multiple-value-setq (visible-file visible-position)
	   (buffer-position-to-include-file-position start include-map
						     current-file))
	
	 (multiple-value-bind (type file bookmark port)
	     (if* (= start end)
		then :null
	      elseif image-positions
		then (values :image (subseq html-buffer
					    (car image-positions)
					    (cdr image-positions)))
		else (determine-link-type link))
	   (when (and (eq :file type)
		      file
		      (member (pathname-type file)
			      *image-file-types*
			      ;; case insensitive:
			      :test #'equalp))
	     #+ignore
	     (warn "In ~a there is an href link to a ~s file."
		   (enough-namestring visible-file *root-dir*)
		   (pathname-type file))
	     (setq type :image))
	   (ecase type
	     (:syntax-error
	      (note-bad-link
	       current-file
	       "~
Syntax error in link~%   \"~a\"~%   at file position ~d~@[~%   in file ~a~].~%"
	       link
	       visible-position
	       (when (not (eq visible-file current-file))
		 visible-file)))
	     (:null
	      ;; href="", which doesn't make sense
	      (note-bad-link
	       current-file
	       "~
Null link at file position ~d~@[~%   in file ~a~].~%"
	       visible-position
	       (when (not (eq visible-file current-file))
		 visible-file)))
	     (:bookmark
	      (when (null (find-bookmark-definition html-buffer bookmark))
		(note-bad-link
		 current-file
		 "~
Bookmark ~s does not have a definition in
   ~a~%   at file position ~d~@[~%   in file ~a~].~%"
		 bookmark
		 (enough-namestring visible-file *root-dir*)
		 visible-position
		 (when (not (eq visible-file current-file))
		   visible-file))
		(incf status)))
	     (:file
	      (let ((link-file
		     (if* (absolute-pathname-p file)
			then (if* (probe-file file)
				thenret
				else (concat-pathnames *root-dir* file))
			else (merge-pathnames
			      file
			      (path-pathname current-file)))))
		(if* (file-directory-p link-file)
		   then (note-bad-link
			 current-file
			 "~
Link to directory~%   \"~a\"
   at file position ~d~@[~%   in file ~a~].~%"
			 (enough-namestring
			  (enough-namestring link-file *root-dir*)
			  (excl::path-namestring visible-file))
			 visible-position
			 (when (not (eq visible-file current-file))
			   visible-file))
			(incf status)
		 elseif (probe-file link-file)
		   then (when (not (visited-file-p link-file))
			  (note-linked-file link-file))
		 elseif (and *nonexistent-files*
			     (member (file-namestring file)
				     *nonexistent-files*
				     :test #'excl::namestring-compare))
		   thenret
		   else (note-bad-link
			 current-file
			 "~
Link to non-existent file~%   \"~a\"
   at file position ~d~@[~%   in file ~a~].~%"
			 (enough-namestring
			  (enough-namestring link-file *root-dir*)
			  (excl::path-namestring visible-file))
			 visible-position
			 (when (not (eq visible-file current-file))
			   visible-file))
			(incf status))))
	     (:file-bookmark
	      (let ((link-file
		     (if* (absolute-pathname-p file)
			then (if* (probe-file file)
				thenret
				else (concat-pathnames *root-dir* file))
			else (merge-pathnames
			      file
			      (path-pathname current-file)))))
		(if* (file-directory-p link-file)
		   then (note-bad-link
			 current-file
			 "~
Link to directory~%   \"~a\"
   at file position ~d~@[~%   in file ~a~].~%"
			 (enough-namestring
			  (enough-namestring link-file *root-dir*)
			  (excl::path-namestring visible-file))
			 visible-position
			 (when (not (eq visible-file current-file))
			   visible-file))
			(incf status)
		 elseif (probe-file link-file)
		   then (when (not (visited-file-p link-file))
			  (note-linked-file link-file))
			(push (cons bookmark (merge-pathnames current-file))
			      (gethash (namestring link-file)
				       *book-mark-table*))
		 elseif (and *nonexistent-files*
			     (member (file-namestring file)
				     *nonexistent-files*
				     :test #'excl::namestring-compare))
		   thenret
		   else (note-bad-link
			 current-file
			 "~
Link to non-existent file (with bookmark)
   \"~a\"~%   at file position ~d~@[~%   in file ~a~].~%"
			 (enough-namestring
			  (enough-namestring link-file *root-dir*)
			  (excl::path-namestring visible-file))
			 visible-position
			 (when (not (eq visible-file current-file))
			   visible-file))
			(incf status))))
	     (:image
	      (let ((image-file
		     (if* (absolute-pathname-p file)
			then (concat-pathnames *root-dir* file)
			else (merge-pathnames
			      file
			      (path-pathname current-file)))))
		(if* (file-directory-p image-file)
		   then (note-bad-link
			 current-file
			 "~
Image link to directory~%   \"~a\"
   at file position ~d~@[~%   in file ~a~].~%"
			 (enough-namestring
			  (enough-namestring image-file *root-dir*)
			  (excl::path-namestring visible-file))
			 visible-position
			 (when (not (eq visible-file current-file))
			   visible-file))
			(incf status)
		 elseif (probe-file image-file)
		   then (note-image-file
			 (enough-namestring
			  (enough-namestring image-file *root-dir*)
			  (excl::path-namestring visible-file)))
		   else (note-bad-link
			 current-file
			 "~
Link to non-existent image file~%   \"~a\"
   at file position ~d~@[~%   in file ~a~].~%"
			 (enough-namestring
			  (enough-namestring image-file *root-dir*)
			  (excl::path-namestring visible-file))
			 visible-position
			 (when (not (eq visible-file current-file))
			   visible-file))
			(incf status))))
	     (:mailto
	      ;; suppose we could connect to the mail host and check the
	      ;; link...
	      )
	     (:news
	      ;; suppose we could connect to the news server and check the
	      ;; link...
	      )
	     (:ftp
	      ;; suppose we could connect to the ftp site and check the
	      ;; link...
	      )
	     (:http
	      (when *check-all-links*
		(case (http-link-status file ;; the host
					port
					;; the url-tail:
					bookmark)
		  (:bad-host
		   (note-bad-link current-file
				  "~
Couldn't connect to host in link: ~a~%" link)
		   (incf status))
		  (:server-reset
		   (note-bad-link current-file
				  "~
HTTP server for link ~a
    unexpectedly closed the connection.~%"
				  link)
		   (incf status))
		  (:bad-link
		   (note-bad-link current-file
				  "~
Link ~a is bad according to server.~%" link)
		   (incf status))
		  (:ok))))))
	 end))
    (setq current-file (pathname current-file))
    (when (and ignore-files
	       (member (file-namestring current-file) ignore-files
		       :test #'equalp :key #'namestring))
      (return-from check-links-1 0))
    (when (and ignore-directories
	       (member (directory-namestring
			(enough-namestring current-file *root-dir*))
		       ignore-directories
		       :test #'(lambda (a b) (match-regexp b a))
		       :key #'namestring))
      (return-from check-links-1 0))
    (when *print-forms*
      (format t "; ~a:~%" (enough-namestring current-file *root-dir*)))
    (note-file-visited current-file)
;;;    (when (and (pathname-type current-file)
;;;	       (member (pathname-type current-file)
;;;		       *image-file-types*
;;;		       ;; case insensitive:
;;;		       :test #'equalp))
;;;      (warn "href link to a ~s file." (pathname-type current-file)))
    (let ((html-buffer (file-contents current-file))
	  (start 0)
	  bookmarks
	  include-map
	  visible-file
	  visible-position)
      (multiple-value-setq (html-buffer include-map)
	(expand-includes html-buffer current-file))
      (loop
	(multiple-value-bind (match ignore ignore2 link-positions
			      frame-positions comment-positions
			      applet-positions body-positions
			      image-positions ignore3 image2-positions)
	    (match-regexp *link-regexp* html-buffer :return :index
			  :start start :end nil
			  :case-fold t
			  :shortest t)
	  (declare (ignore ignore ignore2 ignore3))
	  (when (null match)
	    (when (setq bookmarks (gethash (namestring current-file)
					   *book-mark-table*))
	      ;; Make sure these bookmarks exist, since they are references
	      ;; in other files.
	      (let (bookmark file)
		(dolist (xbookmark bookmarks)
		  (setq bookmark (car xbookmark))
		  (setq file (cdr xbookmark))
		  (when (null (find-bookmark-definition html-buffer bookmark))
		    (note-bad-link
		     current-file
		     "~
Bookmark ~s does not have a definition in
   ~a
   yet it was referenced from file
   ~a.~%"
		     bookmark
		     (enough-namestring current-file *root-dir*)
		     (enough-namestring file *root-dir*))
		    (incf status))))
	      (setf (gethash (namestring current-file) *book-mark-table*)
		nil))
	    (return))

	  (when (and image2-positions
		     (eq (car image2-positions) (cdr image2-positions)))
	    (setq image2-positions nil))
	  
	  (cond
	   (comment-positions (setq start (cdr comment-positions)))
	   (t
	    (let ((first-positions
		   (or link-positions frame-positions image-positions
		       body-positions applet-positions)))
	      (when first-positions
		(setq start
		  (process-match first-positions html-buffer
				 visible-file visible-position
				 include-map
				 ;; is this an image file?
				 (or body-positions image-positions))))
	      (when image2-positions
		(setq start
		  (process-match image2-positions html-buffer
				 visible-file visible-position
				 include-map
				 image2-positions)))))))))

    (when (and recurse *linked-files*)
      (let (file)
	(loop
	  (when (null *linked-files*) (return))
	  (setq file (pop *linked-files*))
	  (when (and (not (visited-file-p file))
		     (or (null ignore-files)
			 (not (member (file-namestring file)
				      ignore-files
				      :test #'equalp :key #'namestring))))
	    (incf status
		  (check-links-1 file :ignore-files ignore-files
				 :ignore-directories ignore-directories))))))
    status))

(defparameter *analyze-ftp-url* (compile-regexp "^ftp://"))
(defparameter *analyze-http-url*
    (compile-regexp
     #.(concatenate 'simple-string
	 "^http://"
	 "\\([^/:]+\\)"
	 "\\(:\\([0-9]+\\)\\|\\)"
	 "\\(/.*\\|\\)")))
(defparameter *analyze-news-url* (compile-regexp "^news:"))
(defparameter *analyze-mailto-url* (compile-regexp "^mailto:"))
(defparameter *analyze-bookmark* (compile-regexp "\\(.*\\)#\\(.*\\)"))
(defparameter *analyze-syntax-error*
    (compile-regexp "^http:/[^/]"))

(defun determine-link-type (link)
  (when (match-regexp *analyze-syntax-error* link :return nil :case-fold t)
    (return-from determine-link-type :syntax-error))
  (when (match-regexp *analyze-mailto-url* link :return nil :case-fold t)
    (return-from determine-link-type :mailto))
  (when (match-regexp *analyze-news-url* link :return nil :case-fold t)
    (return-from determine-link-type :news))
  (when (match-regexp *analyze-ftp-url* link :return nil :case-fold t)
    (return-from determine-link-type :ftp))
  (multiple-value-bind (match ignore host ignore2 port url-tail)
      (match-regexp *analyze-http-url* link :return :string :case-fold t)
    (declare (ignore ignore ignore2))
    (when match
      (when (string= url-tail "") (setq url-tail "/"))
      (if* (or (null port) (string= "" port))
	 then (setq port *default-http-port*)
	 else (setq port (read-from-string port)))
      (return-from determine-link-type
	(values :http host url-tail port))))
  (multiple-value-bind (match ignore file bookmark)
      (match-regexp *analyze-bookmark* link :return :string :case-fold t)
    (declare (ignore ignore))
    (when match
      (when (string= "" file) (setq file nil))
      (when (string= "" bookmark) (setq bookmark nil))
      (return-from determine-link-type
	(values
	 (if* (and file bookmark)
	    then :file-bookmark
	  elseif file
	    then :file
	    else :bookmark)
	 (alias file)
	 bookmark))))
  ;; all that is left is :file
  (values :file (alias link)))

;; this should be in the regexp module:
(defun regexp-quote (string)
  ;; Return a regexp string which matches exactly STRING and nothing else.
  ;; Just like the emacs function.
  (do* ((special-chars '(#\^ #\[ #\] #\. #\* #\$))
	(from-size (length string))
	(size
	 (+ from-size
	    (count-if #'(lambda (item)
			  (member item special-chars :test #'char=))
		      string)))
	(new (make-string size))
	(from 0 (1+ from))
	(to 0 (1+ to)))
      ((= from from-size) new)
    (let ((from-char (schar string from)))
      (when (member from-char special-chars :test #'char=)
	(setf (schar new to) #\\)
	(incf to))
      (setf (schar new to) from-char))))

(defun find-bookmark-definition (html-buffer bookmark)
  (let ((quoted-bookmark (regexp-quote bookmark)))
    (match-regexp
     (concatenate 'simple-string
       "<a[" '(#\space #\tab #\newline) "]+.*"
       "name=\\(" "\"" quoted-bookmark "\"" "\\|" quoted-bookmark "\\)"
       ".*>")
     html-buffer :return nil :shortest t :case-fold t)))

(defun file-contents (current-file)
  (with-open-file (stream current-file :direction :input)
    (let* ((len (file-length stream))
	   (s (make-string len)))
      (excl::filesys-read-bytes (excl::stream-input-fn stream) s 0 len)
      s)))

(defun note-file-visited (file)
  (setf (gethash (namestring file) *processed-files*) t))

(defun visited-file-p (file)
  (gethash (namestring file) *processed-files*))

(defun note-linked-file (file)
  (pushnew (namestring file)
	   *linked-files*
	   :test #'excl::namestring-compare))

(defun note-image-file (file)
  (setf (gethash (namestring file) *image-files*) t))

(defun image-file-seen-p (file)
  (gethash (namestring file) *image-files*))

(defun note-bad-link (file format-string &rest format-args)
  (flet ((announce-file (file)
	   (when (and (null *print-forms*) (null .printed-file-name.))
	     (format t "~&~%; Processing ~a...~%"
		     (enough-namestring file *root-dir*))
	     (setq .printed-file-name. t))))
    (announce-file file)
    (apply #'format t format-string format-args)))

(defun absolute-pathname-p (pathname)
  (setq pathname (pathname pathname))
  (let (x)
    (and (setq x (pathname-directory pathname))
	 (eq :absolute (car x)))))

(defun concat-pathnames (p1 p2)
  (setq p1 (pathname p1))
  (setq p2 (pathname p2))
  (assert (and (absolute-pathname-p p1)
	       (absolute-pathname-p p2)))
  (let ((np2
	 (make-pathname
	  :host (pathname-host p2)
	  :device (pathname-device p2)
	  :directory (let ((d (pathname-directory p2)))
		       (cons :relative (cdr d)))
	  :name (pathname-name p2)
	  :type (pathname-type p2))))
    (merge-pathnames np2 p1)))

(defparameter *include-regexp*
    (compile-regexp "<!--#include file=\"\\([^\"]+\\)\"-->"))

(defun expand-includes (buffer relative-to
			&aux (included-contents-size 0)
			     (include-directives-size 0)
			     (search-start 0)
			     include-map included-file
			     contents contents-length
			     directive-length
			     match-start match-end)
  ;; return a new BUFFER with the
  ;;    <!--#include file="..."-->
  ;; expanded, as well as a map of the buffer.  The map is a list
  ;; describing the boundaries between included files and original buffer.
  (loop
    (multiple-value-bind (match all-positions match-positions)
	(match-regexp *include-regexp* buffer :return :index
		      :start search-start :case-fold t :shortest t)
      (when (null match) (return (values buffer (nreverse include-map))))
      (setq included-file (merge-pathnames
			   (subseq buffer
				   (car match-positions)
				   (cdr match-positions))
			   (path-pathname relative-to)))
      (setq match-start (car all-positions)
	    match-end (cdr all-positions)
	    contents (file-contents included-file)
	    contents-length (length contents)
	    directive-length (- match-end match-start))
      (incf included-contents-size contents-length)
      (incf include-directives-size directive-length)
      (push (vector match-start
		    :buffer
		    included-file
		    0)
	    include-map)
      (push (vector (+ match-start contents-length)
		    included-file
		    :buffer
		    (+ (- (+ match-start contents-length)
			  included-contents-size)
		       include-directives-size))
	    include-map)
      (setq buffer (concatenate 'simple-string
		     (subseq buffer 0 match-start)
		     contents
		     (subseq buffer match-end)))
      ;; this allows included files to include other files (otherwise we'd
      ;; set search-start to match-start + contents-length).
      (setq search-start match-start))))

(defun buffer-position-to-include-file-position (position include-map
						 &optional buffer-file)
  ;; using the include-map return the real (what the user would see)
  ;; position and the name of the file
  (when (null include-map)
    (return-from buffer-position-to-include-file-position
      (values buffer-file position)))
  (do* ((ll include-map (cdr ll))
	(last nil v)
	(v (car ll) (car ll))
	(pos (elt v 0) (elt v 0)))
      ((or (null ll)
	   (< position pos))
       (when (null last) (setq last (car ll)))
       (let ((pos (elt last 0))
	     (from-file (elt last 1))
	     (to-file (elt last 2))
	     (base (elt last 3)))
	 (values (if* (eq :buffer from-file)
		    then ;; in include file
			 to-file
		    else ;; in buffer
			 buffer-file)
		 (+ base (- position pos)))))))

(defun path-pathname (pathname)
  (setq pathname (pathname pathname))
  (make-pathname :host (pathname-host pathname)
		 :device (pathname-device pathname)
		 :directory (pathname-directory pathname)))

(defun http-link-status (host port url-tail &aux s)
  (unwind-protect
      (let (line)
	(setq s (ignore-errors (acl-socket:make-socket
				:remote-host host :remote-port port)))
	(when (null s) (return-from http-link-status :bad-host))
	(setq line
	  (format nil "GET ~a HTTP/1.0~c~%~c~%"
		  (file-url-sans-bookmark url-tail) #\return #\return))
	(format s "~a" line)
	(force-output s)
	(when (null (setq line (ignore-errors (read-line s))))
	  (return-from http-link-status :server-reset))
	(when (match-regexp *http-ok-response* line
			    :return nil :case-fold t)
	  (return-from http-link-status :ok))
	(return-from http-link-status :bad-link))
    (when s (close s))))

(defparameter *file-url*
    (compile-regexp "\\(.*\\)\\(#.*\\|\\)"))

(defun file-url-sans-bookmark (file-url)
  ;; take "foo#bar" and return "foo"
  (multiple-value-bind (match ignore file)
      (match-regexp *file-url* file-url :return :string :case-fold t)
    (declare (ignore ignore))
    (when match
      (when (string= "" file)
	(error "file portion of ~s is null." file-url))
      file)))

(defvar *link-aliases* (make-hash-table :size 101 :test #'equalp))

(defun initialize-aliases (aliases)
  (setq aliases (aclwin:delimited-string-to-list aliases #\,))
  (clrhash *link-aliases*)
  (dolist (alias aliases)
    (let* ((a-list (aclwin:delimited-string-to-list alias #\=))
	   (name (first a-list))
	   (value (second a-list)))
      (setf (gethash name *link-aliases*) value))))

(defun alias (filename)
  (maphash
   #'(lambda (k v &aux len)
       (when (setq len (substringp k filename))
	 (return-from alias
	   (concatenate 'simple-string v (subseq filename len)))))
   *link-aliases*)
  filename)

(defun substringp (test sequence)
  ;; return length of `test' if `sequence' starts with `test'
  (let ((test-len (length test))
	(seq-len (length sequence)))
    (when (>= test-len seq-len)
      (return-from substringp nil))
    (do* ((i 0 (1+ i)))
	((= i test-len) test-len)
      (when (not (char= (schar test i)
			(schar sequence i)))
	(return nil)))))
